from __future__ import annotations

import dataclasses
from enum import Enum


class ItemType(Enum):
    Ingredient = "Ingredient"
    Meal = "Meal"
    Equipment = "Equipment"


@dataclasses.dataclass
class ItemInfo:
    type: ItemType = dataclasses.field(compare=False)
    name: str = dataclasses.field(compare=True)
    is_cuttable: bool = dataclasses.field(compare=False, default=False)
    steps_needed: int = dataclasses.field(compare=False, default=0)
    finished_progress_name: str = dataclasses.field(compare=False, default="Chopped*")
    needs: list[ItemInfo] = dataclasses.field(compare=False, default_factory=list)
    equipment: ItemInfo | None = dataclasses.field(compare=False, default=None)

    _start_items: list[ItemInfo] = dataclasses.field(
        compare=False, default_factory=list
    )

    def __post_init__(self):
        self.type = ItemType(self.type)

    def get_finished_name(self):
        if "*" in self.finished_progress_name:
            return self.finished_progress_name.replace("*", self.name)
        return self.name

    def create_item(self) -> Item:
        match self.type:
            case ItemType.Ingredient:
                if self.is_cuttable:
                    return CuttableItem(
                        name=self.name,
                        finished=False,
                        steps_needed=self.steps_needed,
                        finished_name=self.get_finished_name(),
                        item_info=self,
                    )
                return Item(name=self.name, item_info=self)
            case ItemType.Equipment:
                return CookingEquipment(name=self.name, item_info=self)
            case ItemType.Meal:
                return Meal(
                    name=self.name,
                    finished=False,
                    steps_needed=self.steps_needed,
                    finished_name=self.get_finished_name(),
                    item_info=self,
                )

    def add_start_item_to_equipment(self, start_item: ItemInfo):
        self._start_items.append(start_item)

    def can_start_meal(self, start_item: Item):
        # TODO check specific order / only specific start items
        return any(
            [start_item.name == s for meal in self._start_items for s in meal.needs]
        )

    def start_meal(self, start_item: Item) -> Item:
        for meal in self._start_items:
            for s in meal.needs:
                if s == start_item.name:
                    return meal.create_item()


class Item:
    """Base class for game items which can be held by a player."""

    def __init__(self, name: str, item_info: ItemInfo, *args, **kwargs):
        self.name = self.__class__.__name__ if name is None else name
        self.item_info = item_info

    def can_combine(self, other):
        return False

    def combine(self, other) -> Item | None:
        pass

    def __repr__(self):
        return f"{self.name}({self.extra_repr})"

    def __eq__(self, other):
        return other and self.name == other.name

    @property
    def extra_repr(self):
        return ""


class ProgressibleItem(Item):
    """Class for items which need to be processed (cut, cooked, ...)"""

    def __init__(
        self,
        finished: bool = False,
        steps_needed: int = 500,
        finished_name: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.progressed_steps = steps_needed if finished else 0
        self.steps_needed = steps_needed
        self.finished = finished
        self.finished_name = (
            f"Chopped{self.name}" if finished_name is None else finished_name
        )

    def progress(self):
        """Progresses the item process as long as it is not finished."""
        if self.progressed_steps >= self.steps_needed:
            self.finished = True
            self.finished_call()
        if not self.finished:
            self.progressed_steps += 1

    def can_progress(self) -> bool:
        return True

    def finished_call(self):
        self.name = self.finished_name

    def reset(self):
        self.finished = False
        self.progressed_steps = 0

    def __repr__(self):
        if not self.steps_needed or self.finished:
            return f"{self.name}({self.extra_repr})"
        else:
            return f"{self.name}(progress={int(self.progressed_steps / self.steps_needed * 100)}%,{self.extra_repr})"


class CuttableItem(ProgressibleItem):
    """Class of item which can be processed by the cutting board."""


class CookingEquipment(Item):
    def __init__(self, content: Meal = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content = content

    def can_combine(self, other):
        if self.content is None:
            if isinstance(other, CookingEquipment):
                return other.can_release_content()
            # TODO check other is start of a meal, create meal
            if isinstance(other, Meal) and self.name == "Plate":
                return not other.steps_needed or other.finished
            return self.item_info.can_start_meal(other)
        return self.content.can_combine(other)

    def combine(self, other):
        if self.content is None:
            if isinstance(other, CookingEquipment):
                self.content = other.release()
                return other
            if isinstance(other, Meal) and self.name == "Plate":
                self.content = other
                return
            # find starting meal for other
            self.content = self.item_info.start_meal(other)
        self.content.combine(other)

    def can_progress(self, counter_type="Stove") -> bool:
        return (
            self.content
            and isinstance(self.content, ProgressibleItem)
            and self.content.can_progress()
        )

    def progress(self):
        self.content.progress()

    def can_release_content(self) -> bool:
        return (
            self.content
            and isinstance(self.content, ProgressibleItem)
            and self.content.finished
        )

    def release(self):
        content = self.content
        self.content = None
        return content

    @property
    def extra_repr(self):
        return self.content


class Meal(ProgressibleItem):
    def __init__(self, parts=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parts = [] if parts is None else parts
        # self.rules ...

    def can_combine(self, other) -> bool:
        if other and not self.finished:
            satisfied = [False for _ in range(len(self.parts))]
            for n in self.item_info.needs:
                for i, p in enumerate(self.parts):
                    if not satisfied[i] and p.name == n:
                        satisfied[i] = True
                        break
                else:
                    if n == other.name:
                        return True
        return False

    def combine(self, other):
        self.parts.append(other)

    def can_progress(self) -> bool:
        return self.item_info.steps_needed and len(self.item_info.needs) == len(
            self.parts
        )

    def finished_call(self):
        super().finished_call()

    @property
    def extra_repr(self):
        return self.parts
