import time

import numpy as np
import pytest

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.counters import Counter, CuttingBoard
from overcooked_simulator.game_items import CuttableItem
from overcooked_simulator.overcooked_environment import Action
from overcooked_simulator.simulation_runner import Simulator

layouts_folder = ROOT_DIR / "game_content" / "layouts"


# TODO: TESTs are in absolute pixel coordinates still.


@pytest.fixture(autouse=True)
def test_file_availability():
    assert layouts_folder.is_dir(), "layouts folder does not exists"
    assert (
        layouts_folder / "basic.layout"
    ).is_file(), "basic layouts file does not exists"


def test_player_registration():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        200,
    )

    sim.register_player("player1", np.array([0, 0]))

    assert len(sim.env.players) != 0, "Wrong number of players"
    assert len(sim.env.players) == 1, "Wrong number of players"

    sim.register_player("player2", np.array([100, 100]))

    assert len(sim.env.players) == 2, "Wrong number of players"

    sim.register_player("player2", np.array([100, 100]))  # same player name
    assert len(sim.env.players) == 2, "Wrong number of players"

    sim.start()
    sim.stop()


def test_simulator_frequency():
    class TestEnv:
        def __init__(self):
            self.c = 0

        def step(self):
            self.c += 1

    frequency = 2000
    running_time_seconds = 2

    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        frequency,
    )
    sim.env = TestEnv()  # Overwrite environment with a simple counting env

    sim.start()
    time.sleep(running_time_seconds)
    sim.stop()

    print(sim.env.c)
    accepted_tolerance = 0.02
    lower = frequency * running_time_seconds * (1 - accepted_tolerance)
    upper = frequency * running_time_seconds * (1 + accepted_tolerance)
    assert lower < sim.env.c < upper, "Timing error in the environment at 1000hz"


def test_movement():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        200,
    )
    player_name = "p1"
    start_pos = np.array([50, 50])
    sim.register_player(player_name, start_pos)
    move_direction = np.array([1, 0])
    move_action = Action(player_name, "movement", move_direction)
    do_moves_number = 6
    for i in range(do_moves_number):
        sim.enter_action(move_action)

    expected = start_pos + do_moves_number * (
        move_direction * sim.env.players[player_name].move_dist
    )

    assert (
        np.linalg.norm(expected - sim.env.players[player_name].pos) == 0
    ), "Should be here?"


def test_collision_detection():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        200,
    )
    counter_pos = np.array([50, 100])
    counter = Counter(counter_pos)
    sim.env.counters = [counter]
    sim.register_player("p1", np.array([50, 50]))
    sim.register_player("p2", np.array([50, 200]))  # same player name
    player1 = sim.env.players["p1"]
    player2 = sim.env.players["p2"]

    sim.start()
    assert not sim.env.detect_collision_counters(player1), "Should not collide"
    assert not sim.env.detect_player_collision(player1), "Should not collide yet."

    assert not sim.env.detect_collision(player1), "Does not collide yet."

    player1.move_abs(counter_pos)
    assert sim.env.detect_collision_counters(
        player1
    ), "Player and counter at same pos. Not detected."
    player2.move_abs(counter_pos)
    assert sim.env.detect_player_collision(
        player1
    ), "Players at same pos. Not detected."

    player1.move_abs(np.array([0, 0]))
    assert sim.env.detect_collision_world_bounds(
        player1
    ), "Player collides with world bounds."
    sim.stop()


def test_player_reach():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        200,
    )

    counter_pos = np.array([100, 100])
    counter = Counter(counter_pos)
    sim.env.counters = [counter]
    sim.register_player("p1", np.array([100, 200]))
    player = sim.env.players["p1"]
    assert not player.can_reach(counter), "Player is too far away."

    do_moves_number = 30
    for i in range(do_moves_number):
        move_action = Action("p1", "movement", np.array([0, -1]))
        sim.enter_action(move_action)
    assert player.can_reach(counter), "Player can reach counter?"


def test_pickup():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        200,
    )

    counter_pos = np.array([100, 100])
    counter = Counter(counter_pos)
    counter.occupied_by = CuttableItem(name="Tomato", item_info=None)
    sim.env.counters = [counter]

    sim.register_player("p1", np.array([100, 160]))
    player = sim.env.players["p1"]

    move_down = Action("p1", "movement", np.array([0, -1]))
    move_up = Action("p1", "movement", np.array([0, 1]))
    pick = Action("p1", "pickup", "pickup")

    sim.enter_action(move_down)
    assert player.can_reach(counter), "Player can reach counter?"

    sim.enter_action(pick)
    assert player.holding is not None, "Player should have picked up tomato."
    assert player.holding.name == "Tomato", "Should be tomato."

    sim.enter_action(move_up)
    sim.enter_action(move_up)
    sim.enter_action(move_up)
    sim.enter_action(move_up)

    sim.enter_action(pick)
    assert (
        player.holding is not None
    ), "Player should be too far away to put tomato down."

    sim.enter_action(move_down)
    sim.enter_action(move_down)
    sim.enter_action(move_down)
    sim.enter_action(move_down)

    sim.enter_action(pick)

    assert player.holding is None, "Player should have put tomato down."
    assert (
        counter.occupied_by is not None and counter.occupied_by.name == "Tomato"
    ), "Tomato should be here now."


def test_processing():
    sim_frequency = 1000
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        sim_frequency,
    )
    sim.start()

    counter_pos = np.array([100, 100])
    counter = CuttingBoard(counter_pos)
    sim.env.counters = [counter]

    tomato = CuttableItem(name="Tomato", item_info=None)
    sim.register_player("p1", np.array([100, 150]))
    player = sim.env.players["p1"]
    player.holding = tomato

    move = Action("p1", "movement", np.array([0, -1]))
    pick = Action("p1", "pickup", "pickup")

    sim.enter_action(move)
    sim.enter_action(pick)

    hold_down = Action("p1", "interact", "keydown")
    sim.enter_action(hold_down)

    assert not tomato.finished, "Tomato is not finished yet."

    sleep_time = (tomato.steps_needed / sim_frequency) + 0.1
    time.sleep(sleep_time)

    assert tomato.finished, "Tomato should be finished."

    button_up = Action("p1", "interact", "keyup")
    sim.enter_action(button_up)

    sim.stop()
